;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;; *************************************************************************
;;;; FILE IDENTIFICATION
;;;;
;;;; Name:          attrib-class.lisp
;;;; Purpose:       Defines metaclass allowing use of attributes on slots
;;;; Programmer:    Kevin M. Rosenberg
;;;; Date Started:  Apr 2000
;;;;
;;;; $Id: attrib-class.lisp,v 1.9 2003/04/29 03:50:42 kevin Exp $
;;;;
;;;; This file, part of KMRCL, is Copyright (c) 2002 by Kevin M. Rosenberg
;;;;
;;;; KMRCL users are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser GNU Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.
;;;; *************************************************************************

;;;; Defines a metaclass that allows the use of attributes (or subslots)
;;;; on slots. Based on example in AMOP, but modified to use ACL's MOP.

(in-package :kmrcl)

(defclass attributes-dsd (kmr-mop:standard-direct-slot-definition)
  ((attributes :initarg :attributes :initform nil
	       :accessor dsd-attributes)))

(defclass attributes-esd (kmr-mop:standard-effective-slot-definition)
  ((attributes :initarg :attributes :initform nil 
	       :accessor esd-attributes)))


(defclass attributes-class (kmr-mop:standard-class)
  ()
  (:documentation "metaclass that implements attributes on slots. Based
on example from AMOP"))


#+(or cmu scl sbcl)
(defmethod kmr-mop:validate-superclass ((class attributes-class)
					(superclass kmr-mop:standard-class))
  t)

(defmethod kmr-mop:direct-slot-definition-class ((cl attributes-class) #+kmr-normal-dsdc &rest initargs)
  (declare (ignore initargs))
  (kmr-mop:find-class 'attributes-dsd))

(defmethod kmr-mop:compute-effective-slot-definition :around
    ((cl attributes-class) #+kmr-normal-cesd name dsds)
  #+kmr-normal-cesd (declare (ignore name))
  (apply
   #'make-instance 'attributes-esd 
   :attributes (remove-duplicates (mapappend #'dsd-attributes dsds))
   (kmr-mop:compute-effective-slot-definition-initargs cl dsds))
  )

#+ignore
(defmethod kmr-mop:compute-effective-slot-definition :around
    ((cl attributes-class) #+kmr-named-cesd name dsds)
  #+kmr-named-cesd (declare (ignore name))
  (let ((normal-slot (call-next-method)))
    (setf (esd-attributes normal-slot)
      (remove-duplicates
       (mapappend #'esd-attributes dsds)))
    normal-slot))


(defmethod kmr-mop:compute-slots ((class attributes-class))
  (let* ((normal-slots (call-next-method))
	 (alist
	  (mapcar
	   #'(lambda (slot)
	       (let ((attr-list (mapcar #'(lambda (attr) (cons attr nil))
					(esd-attributes slot))))
		 (when attr-list
		   (cons (kmr-mop:slot-definition-name slot) attr-list))))
	   normal-slots)))
    (setq alist (delete nil alist))
    (cons (make-instance 'kmr-mop:standard-effective-slot-definition
	    :name 'all-attributes
	    :initform `',alist
	    :initfunction #'(lambda () alist))
	  normal-slots)))
  
(defun slot-attribute (instance slot-name attribute)
  (cdr (slot-attribute-bucket instance slot-name attribute)))

(defun (setf slot-attribute) (new-value instance slot-name attribute)
  (setf (cdr (slot-attribute-bucket instance slot-name attribute))
    new-value))

(defun slot-attribute-bucket (instance slot-name attribute)
  (let* ((all-buckets (slot-value instance 'all-attributes))
	 (slot-bucket (assoc slot-name all-buckets)))
    (unless slot-bucket
      (error "The slot named ~S of ~S has no attributes."
	     slot-name instance))
    (let ((attr-bucket (assoc attribute (cdr slot-bucket))))
      (unless attr-bucket
	(error "The slot named ~S of ~S has no attributes named ~S."
	       slot-name instance attribute))
      attr-bucket)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (export '(attributes-class slot-attributes)))

